/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <system_info.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <aul.h>
#include <cert-svc/ccert.h>
#include <cert-svc/cinstance.h>
#include <cynara-session.h>
#include <cynara-client.h>
#include <cynara-creds-gdbus.h>
#include <tzplatform_config.h>

#include "alarm-manager-util.h"

#define ALARMMGR_LOG_BUFFER_SIZE	10000
#define ALARMMGR_LOG_BUFFER_STRING_SIZE	400
#define ALARMMGR_LOG_DIR_PATH	"/var/log/appfw/alarmmgr_log/"
#define ALARMMGR_LOG_FILE_PATH	"/var/log/appfw/alarmmgr_log/alarmmgr.log"
static int log_index = 0;
static int log_fd = 0;

static int __bg_category_func(const char *name, void *user_data)
{
	bg_category_cb_info_t *info = (bg_category_cb_info_t *)user_data;
	LOGD("appid[%s], bg name = %s", info->appid, name);
	if (name && strncmp("enable", name, strlen(name)) &&
			strncmp("disable", name, strlen(name))) {
		info->has_bg = true;
		return -1;
	}

	return 0;
}

profile_t _get_profile()
{
	static profile_t saved = PROFILE_UNKNOWN;
	char *profileName;
	int r;

	if (__builtin_expect(saved != PROFILE_UNKNOWN, 1))
		return saved;

	r = system_info_get_platform_string("http://tizen.org/feature/profile",
			&profileName);
	if (r != SYSTEM_INFO_ERROR_NONE) {
		LOGD("Failed to get profile info. error(%d)", r);
		return saved;
	}

	switch (*profileName) {
	case 'm':
	case 'M':
		saved = PROFILE_MOBILE;
		break;
	case 'w':
	case 'W':
		saved = PROFILE_WEARABLE;
		break;
	case 't':
	case 'T':
		saved = PROFILE_TV;
		break;
	case 'i':
	case 'I':
		saved = PROFILE_IVI;
		break;
	default: // common or unknown ==> ALL ARE COMMON.
		saved = PROFILE_COMMON;
	}
	free(profileName);

	return saved;
}

int _pkg_is_global(const char *callee_pkgid, uid_t uid)
{
	int retval;
	int return_code = ERR_ALARM_SYSTEM_FAIL;
	pkgmgrinfo_pkginfo_h handle;

	retval = pkgmgrinfo_pkginfo_get_usr_pkginfo(callee_pkgid, uid, &handle);
	if (retval != PMINFO_R_OK) {
		LOGE("Failed to get pkginfo\n");
		return_code = ERR_ALARM_INVALID_ID;
	} else {
		bool is_global = 0;
		retval = pkgmgrinfo_pkginfo_is_global(handle, &is_global);
		if (retval == PMINFO_R_OK && is_global)
			return_code = ALARMMGR_RESULT_SUCCESS;
		else if (retval == PMINFO_R_OK && !is_global)
			return_code = ERR_ALARM_NOT_PERMITTED_APP;

		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
	}

	return return_code;
}

int _cynara_check(GDBusMethodInvocation *invocation, const char *privilege,
		pid_t pid)
{
	int ret = 0;
	char *user = NULL;
	char *client = NULL;
	char *client_session = NULL;
	cynara *p_cynara = NULL;
	const char *sender_unique_name;
	GDBusConnection *connection;

	connection = g_dbus_method_invocation_get_connection(invocation);
	sender_unique_name = g_dbus_method_invocation_get_sender(invocation);

	ret = cynara_initialize(&p_cynara, NULL);
	if (ret != CYNARA_API_SUCCESS) {
		LOGE("cynara_initialize() failed : %d", ret);
		ret = ERR_ALARM_SYSTEM_FAIL;
		goto cynara_out;
	}

	ret = cynara_creds_gdbus_get_user(connection, sender_unique_name,
			USER_METHOD_DEFAULT, &user);
	if (ret != CYNARA_API_SUCCESS) {
		LOGE("cynara_creds_gdbus_get_user() failed : %d", ret);
		ret = ERR_ALARM_SYSTEM_FAIL;
		goto cynara_out;
	}

	ret = cynara_creds_gdbus_get_client(connection, sender_unique_name,
			CLIENT_METHOD_DEFAULT, &client);
	if (ret != CYNARA_API_SUCCESS) {
		LOGE("cynara_creds_gdbus_get_client() failed : %d", ret);
		ret = ERR_ALARM_SYSTEM_FAIL;
		goto cynara_out;
	}

	LOGD("user :%s , client :%s ,unique_name : %s, pid() : %d",
			user, client, sender_unique_name, pid);

	client_session = cynara_session_from_pid(pid);
	if (!client_session) {
		LOGE("cynara_session_from_pid() failed : %d", ret);
		ret = ERR_ALARM_SYSTEM_FAIL;
		goto cynara_out;
	}

	ret = cynara_check(p_cynara, client, client_session, user,
			privilege);
	if (ret == CYNARA_API_ACCESS_ALLOWED) {
		LOGD("CYNARA_ACCESS_ALLOWED");
		ret = ALARMMGR_RESULT_SUCCESS;
	}
	else {
		LOGD("CYNARA_NOT_ALLOWED [%d]", ret);
		ret = ERR_ALARM_NOT_PERMITTED_APP;
	}

cynara_out:
	if (client_session)
		g_free(client_session);
	if (client)
		g_free(client);
	if (user)
		g_free(user);
	if (p_cynara)
		cynara_finish(p_cynara);

	return ret;
}

int _get_pid_from_appid(const char *app_id, uid_t uid)
{
	int pid;
	pid = aul_app_get_pid_for_uid(app_id, uid);
	if (pid < 0) {
		LOGE("aul_app_get_pid_for_uid() is failed. appid %s may not be app. uid : [%d]", app_id, uid);
		return 0;
	}

	LOGD("app_id:%s [pid:%d]", app_id, pid);
	return pid;
}

char *_get_pkgid_by_appid(const char *app_id, uid_t uid)
{
	pkgmgrinfo_pkginfo_h handle;
	char *pkgid = NULL;
	char *temp;

	if (pkgmgrinfo_appinfo_get_usr_appinfo(app_id, uid, &handle) == PMINFO_R_OK) {
		if (pkgmgrinfo_appinfo_get_pkgid(handle, &temp) == PMINFO_R_OK) {
			if (temp)
				pkgid = strdup(temp);
		}
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
	}

	return pkgid;
}

int _is_ui_app(const char *appid, uid_t uid)
{
	int ret = 0;
	int found = 0;
	char *component = NULL;
	pkgmgrinfo_appinfo_h appinfo_h = NULL;

	if (appid == NULL)
		return 0;

	ret = pkgmgrinfo_appinfo_get_usr_appinfo(appid, uid, &appinfo_h);
	if (ret < 0)
		return 0;

	ret = pkgmgrinfo_appinfo_get_component_type(appinfo_h, &component);
	if (ret == 0 && component != NULL && strncmp(component, "uiapp", 5) == 0)
		found = 1;

	if (appinfo_h)
		pkgmgrinfo_appinfo_destroy_appinfo(appinfo_h);

	return found;
}

bool _is_app(const char *appid, uid_t uid)
{
	pkgmgrinfo_appinfo_h appinfo_h = NULL;
	int ret;

	ret = pkgmgrinfo_appinfo_get_usr_appinfo(appid, uid, &appinfo_h);
	if (ret != PMINFO_R_OK)
		return false;

	if (appinfo_h)
		pkgmgrinfo_appinfo_destroy_appinfo(appinfo_h);

	return true;
}

int _compare_api_version(int *result, int pid, uid_t uid)
{
	int ret = 0;
	pkgmgrinfo_pkginfo_h pkginfo = NULL;
	char pkgid[MAX_PKG_ID_LEN] = {0, };
	char *pkg_version;

	if (aul_app_get_pkgid_bypid_for_uid(pid, pkgid, sizeof(pkgid), uid) != AUL_R_OK) {
		LOGE("aul_app_get_pkgid_bypid() is failed. PID %d may not be app.", getpid());
	} else {
		ret = pkgmgrinfo_pkginfo_get_usr_pkginfo(pkgid, uid, &pkginfo);
		if (ret != PMINFO_R_OK) {
			LOGE("Failed to get pkginfo\n");
		} else {
			ret = pkgmgrinfo_pkginfo_get_api_version(pkginfo, &pkg_version);
			if (ret != PMINFO_R_OK)
				LOGE("Failed to check api version [%d]\n", ret);
			*result = strverscmp(pkg_version, "2.4");
			pkgmgrinfo_pkginfo_destroy_pkginfo(pkginfo);
		}
	}
	return ret;
}

bool _permit_by_config(pkgmgrinfo_appinfo_h handle, uid_t uid)
{
	if (access(tzplatform_mkpath(TZ_SYS_RO_SHARE,
					"alarm-manager/alarm-config-service-restricted"), F_OK) == 0) {
		LOGD("This profile restrict alarms for service applications\n");
		return false;
	}

	if (access(tzplatform_mkpath(TZ_SYS_RO_SHARE,
					"alarm-manager/alarm-config-platform-service-permitted"), F_OK) == 0) {
		LOGD("This profile permit alarm for service applications which has platform cert\n");
		char *pkgid;
		int r;
		const char *cert_value;
		pkgmgrinfo_certinfo_h certinfo;
		CertSvcInstance instance;
		CertSvcCertificate certificate;
		CertSvcVisibility visibility = CERTSVC_VISIBILITY_PUBLIC;

		r = pkgmgrinfo_appinfo_get_pkgid(handle, &pkgid);
		if (r != PMINFO_R_OK) {
			LOGE("Failed to get certinfo pkgid");
			return false;
		}

		r = pkgmgrinfo_pkginfo_create_certinfo(&certinfo);
		if (r != PMINFO_R_OK) {
			LOGE("Failed to create certinfo");
			return false;
		}

		r = pkgmgrinfo_pkginfo_load_certinfo(pkgid, certinfo, uid);
		if (r != PMINFO_R_OK) {
			LOGE("Failed to load certinfo");
			pkgmgrinfo_pkginfo_destroy_certinfo(certinfo);
			return false;
		}

		r = pkgmgrinfo_pkginfo_get_cert_value(certinfo,
				PMINFO_DISTRIBUTOR_ROOT_CERT, &cert_value);
		if (r != PMINFO_R_OK || cert_value == NULL) {
			LOGE("Failed to get cert value");
			pkgmgrinfo_pkginfo_destroy_certinfo(certinfo);
			return false;
		}

		r = certsvc_instance_new(&instance);
		if (r != CERTSVC_SUCCESS) {
			LOGE("certsvc_instance_new() is failed.");
			pkgmgrinfo_pkginfo_destroy_certinfo(certinfo);
			return false;
		}

		r = certsvc_certificate_new_from_memory(instance,
				(const unsigned char *)cert_value,
				strlen(cert_value),
				CERTSVC_FORM_DER_BASE64,
				&certificate);
		if (r != CERTSVC_SUCCESS) {
			LOGE("certsvc_certificate_new_from_memory() is failed.");
			pkgmgrinfo_pkginfo_destroy_certinfo(certinfo);
			certsvc_instance_free(instance);
			return false;
		}

		r = certsvc_certificate_get_visibility(certificate, &visibility);
		if (r != CERTSVC_SUCCESS)
			LOGE("certsvc_certificate_get_visibility() is failed.");

		pkgmgrinfo_pkginfo_destroy_certinfo(certinfo);
		certsvc_instance_free(instance);
		certsvc_certificate_free(certificate);

		LOGW("visibility is %d", visibility);
		if (visibility & CERTSVC_VISIBILITY_PLATFORM) {
			return true;
		}
	}

	if (access(tzplatform_mkpath(TZ_SYS_RO_SHARE,
					"alarm-manager/alarm-config-all-service-permitted"), F_OK) == 0) {
		LOGD("This profile permit alarms for all service applications\n");
		return true;
	}

	return false;
}

bool _is_permitted(const char *app_id, int alarm_type, uid_t uid)
{
	pkgmgrinfo_appinfo_h handle = NULL;
	int ret;
	bool _return = false;

	if (app_id == NULL) {
		LOGE("app_id is NULL. Only expicit launch is permitted\n");
		return false;
	}

	ret = pkgmgrinfo_appinfo_get_usr_appinfo(app_id, uid, &handle);
	if (ret != PMINFO_R_OK) {
		LOGE("Failed to get appinfo [%s]\n", app_id);
	} else {
		char *app_type = NULL;
		ret = pkgmgrinfo_appinfo_get_component_type(handle, &app_type);
		if (app_type && strcmp("uiapp", app_type) == 0) {
			if (alarm_type & ALARM_TYPE_EXACT_SERVICE_APP) {
				LOGE("[%s] is ui application.\
						But alarm_type is ALARM_TYPE_EXACT_SERVICE_APP.\
						soit is not allowed", app_id);
				_return = false;
				goto out;
			}
			LOGD("[%s] is ui application. It is allowed", app_id);
			_return = true;
			goto out;
		} else if (app_type && strcmp("svcapp", app_type) == 0) {
			LOGD("[%s] is service application.", app_id);

			if (_permit_by_config(handle, uid)) {
				LOGD("service applications are allowed");
				_return = true;
				goto out;
			}

			bg_category_cb_info_t info = {
				.appid = app_id,
				.has_bg = false
			};

			if (alarm_type & ALARM_TYPE_INEXACT || alarm_type & ALARM_TYPE_EXACT_SERVICE_APP) {
				ret = pkgmgrinfo_appinfo_foreach_background_category(handle, __bg_category_func, &info);
				if (ret == PMINFO_R_OK && info.has_bg) {
					LOGD("[%s] has background categories. It is allowed", app_id);
					_return = true;
					goto out;
				} else {
					LOGE("Failed to foreach background category. [%s] is not allowed", app_id);
				}
			}
		}
	}

out:
	if (handle)
		pkgmgrinfo_appinfo_destroy_appinfo(handle);

	return _return;
}

void _initialize_module_log(void)
{
#ifdef _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG
	char buf[1024];

	if (access(ALARMMGR_LOG_DIR_PATH, F_OK)) {
		LOGE("Not exist(%s)", ALARMMGR_LOG_DIR_PATH);

		if (mkdir(ALARMMGR_LOG_DIR_PATH, (S_IRUSR | S_IWUSR | S_IXUSR |
					S_IRGRP | S_IWGRP | S_IXGRP))) {
			LOGE("Failed to create directory(%s). errno(%d)",
					ALARMMGR_LOG_DIR_PATH, errno);
			return;
		}
	}

	log_fd = open(ALARMMGR_LOG_FILE_PATH, O_CREAT | O_WRONLY, 0644);
	if (log_fd == -1) {
		LOGE("Opening the file for alarmmgr log is failed. err: %s", strerror_r(errno, buf, sizeof(buf)));
		return;
	}

	int offset = lseek(log_fd, 0, SEEK_END);
	if (offset != 0) {
		log_index = (int)(offset / ALARMMGR_LOG_BUFFER_STRING_SIZE);
		if (log_index >= ALARMMGR_LOG_BUFFER_SIZE) {
			log_index = 0;
			lseek(log_fd, 0, SEEK_SET);
		}
	}
#endif /* _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG */
	return;
}

void _save_module_log(const char *tag, const char *message)
{
#ifdef _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG
	char buffer[ALARMMGR_LOG_BUFFER_STRING_SIZE] = {0,};
	time_t now;
	char buf[1024];
	char *time_str;

	if (log_fd == -1) {
		LOGE("The file is not ready.");
		return;
	}

	if (log_index != 0)
		lseek(log_fd, 0, SEEK_CUR);
	else
		lseek(log_fd, 0, SEEK_SET);

	time(&now);
	time_str = ctime(&now);
	if (time_str)
		time_str[strlen(time_str) - 1] = '\0'; /* to avoid new line */

	snprintf(buffer, ALARMMGR_LOG_BUFFER_STRING_SIZE, "[%-6d]%s\t%s\t : %ld %s\n", log_index, tag, message, now, time_str);

	SECURE_LOGW("%s", buffer);
	int ret = write(log_fd, buffer, strlen(buffer));
	if (ret < 0) {
		LOGE("Writing the alarmmgr log is failed. err: %s", strerror_r(errno, buf, sizeof(buf)));
		return;
	}

	if (++log_index >= ALARMMGR_LOG_BUFFER_SIZE)
		log_index = 0;
#endif /* _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG */
	return;
}

void _save_alarm_info_log(const char *tag, __alarm_info_t *info)
{
#ifdef _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG
	char log_message[ALARMMGR_LOG_MESSAGE_SIZE] = {0,};
	char *due_time_str = ctime(&(info->due_time));
	if (due_time_str)
		due_time_str[strlen(due_time_str) - 1] = '\0'; /* to avoid new line */

	char *timebuf = strdup(ctime(&info->due_time));
	if (timebuf)
		timebuf[strlen(timebuf) - 1] = '\0'; /* to avoid new line */

	snprintf(log_message, sizeof(log_message),
		"alarmID: %d, uid: %d, caller : %s -> callee : %s, "
		"dst_name %s, dst_mod %s, "
		"type 0x%x, mode %d, "
		"interval:%ld, duetime: %ld %s",
		info->alarm_id, info->uid, info->app_unique_name, info->callee_pkgid,
		info->dst_service_name, info->dst_service_name_mod,
		info->alarm_info.alarm_type, info->alarm_info.mode.repeat,
		info->alarm_info.mode.u_interval.interval, info->due_time, timebuf);

	if (timebuf)
		free(timebuf);

	_save_module_log(tag, log_message);
#endif /* _APPFW_FEATURE_ALARM_MANAGER_MODULE_LOG */
	return;
}

